@@ -15,7 +15,7 @@
 
 /*
    Modified by Timothy Mann, 1996
-   Last modified on Sat Dec 12 11:47:51 PST 1998 by mann
+   Last modified on Tue May  1 17:52:20 PDT 2001 by mann
 */
 
 #include "z80.h"
@@ -55,11 +55,11 @@ static char help_message[] =
 \n\
 Running:\n\
     run\n\
-        Reset the Z-80 and commence execution at address 0000.\n\
+        Hard reset the Z-80 and devices and commence execution.\n\
     cont\n\
         Continue execution.\n\
     step\n\
-        Execute one instruction, disallowing interrupts.\n\
+        Execute one instruction, disallowing all interrupts.\n\
     stepint\n\
         Execute one instruction, allowing an interrupt afterwards.\n\
     next\n\
@@ -68,7 +68,11 @@ Running:\n\
         until the return.  Interrupts are always allowed inside the call,\n\
         but only the nextint form allows an interrupt afterwards.\n\
     reset\n\
-        Reset the Z-80.\n\
+        Hard reset the Z-80 and devices.\n\
+    softreset\n\
+        Press the system reset button.  On Model I/III, softreset resets the\n\
+        devices and posts a nonmaskable interrupt to the CPU; on Model 4/4P,\n\
+        softreset is the same as hard reset.\n\
     timeroff\n\
     timeron\n\
         Disable/enable the emulated TRS-80 real time clock interrupt.\n\
@@ -100,6 +104,7 @@ Traps:\n\
     delete *\n\
         Delete trap n, or all traps.\n\
     stop at <address>\n\
+    break <address>\n\
         Set a breakpoint at the specified hex address.\n\
     trace <address>\n\
         Set a trap to trace execution at the specified hex address.\n\
@@ -115,7 +120,7 @@ Miscellaneous:\n\
     ?\n\
         Print this message.\n\
     quit\n\
-        Exit the zbx.\n";
+        Exit from xtrs.\n";
 
 static struct
 {
@@ -263,11 +268,13 @@ void debug_print_registers()
 static void signal_handler()
 {
     stop_signaled = 1;
+    if (trs_continuous > 0) trs_continuous = 0;
 }
 
 void trs_debug()
 {
     stop_signaled = 1;
+    if (trs_continuous > 0) trs_continuous = 0;
 }
 
 void debug_init()
@@ -322,6 +329,7 @@ static void debug_run()
 {
     void (*old_signal_handler)();
     Uchar t;
+    int continuous;
 
     /* catch control-c signal */
     old_signal_handler = signal(SIGINT, signal_handler);
@@ -350,11 +358,10 @@ static void debug_run()
 	
 	if(print_instructions) disassemble(REG_PC);
 	
-	if(z80_run(0))
-	{
-	    /* halt instruction */
-	    printf("Halt instruction executed.\n");
-	    stop_signaled = 1;
+	continuous = (!print_instructions && num_traps == 0);
+	if (z80_run(continuous)) {
+	  printf("emt_debug instruction executed.\n");
+	  stop_signaled = 1;
 	}
 
 	t = traps[REG_PC];
@@ -386,9 +393,7 @@ void debug_shell()
     if (!home) home = ".";
     sprintf (history_file, "%s/.zbx-history", home);
     read_history(history_file);
-#endif READLINE
-
-    z80_reset();
+#endif
 
     while(!done)
     {
@@ -404,14 +409,6 @@ void debug_shell()
 	/*
 	 * Use the way cool gnu readline() utility.  Get completion,
 	 * history, way way cool.
-         *
-	 * ... but unfortunately, broken.  DO NOT USE THIS CODE.
-	 * Currently there is a bug in GNU readline() that makes it
-	 * incompatible with my timer interrupt implementation.
-	 * Readline thinks that it is always supposed to clean up its
-	 * state and exit when a SIGALRM comes in, but in fact, I
-	 * handle the signal and continue the read system call.  I've
-	 * reported this bug to readline's maintainer.  --tpm
          */
         {
 
@@ -476,13 +473,14 @@ void debug_shell()
 	    else if(!strcmp(command, "list"))
 	    {
 		int x, y;
-		Ushort start, end = 0;
+		Ushort start, old_start;
+		int bytes = 0;
 		int lines = 0;
 
 		if(sscanf(input, "list %x , %x", &x, &y) == 2)
 		{
 		    start = x;
-		    end = y;
+		    bytes = (y - x) & 0xffff;
 		}
 		else if(sscanf(input, "list %x", &x) == 1)
 		{
@@ -504,9 +502,9 @@ void debug_shell()
 		}
 		else
 		{
-		    while(start <= end)
-		    {
-			start = disassemble(start);
+		    while (bytes >= 0) {
+			start = disassemble(old_start = start);
+			bytes -= (start - old_start) & 0xffff;
 		    }
 		}
 	    }
@@ -570,13 +568,18 @@ void debug_shell()
 	    }
 	    else if(!strcmp(command, "reset"))
 	    {
-		printf("Resetting.");
-		z80_reset();
+		printf("Performing hard reset.");
+		trs_reset(1);
+	    }
+	    else if(!strcmp(command, "softreset"))
+	    {
+		printf("Pressing reset button.");
+		trs_reset(0);
 	    }
 	    else if(!strcmp(command, "run"))
 	    {
-		printf("Running.\n");
-		z80_reset();
+		printf("Performing hard reset and running.\n");
+		trs_reset(1);
 		debug_run();
 	    }
 	    else if(!strcmp(command, "status"))
@@ -653,11 +656,12 @@ void debug_shell()
 	    {
 		z80_run(0);
 	    }
-	    else if(!strcmp(command, "stop"))
+	    else if(!strcmp(command, "stop") || !strcmp(command, "break"))
 	    {
 		int address;
 
-		if(sscanf(input, "stop at %x", &address) != 1)
+		if(sscanf(input, "stop at %x", &address) != 1 &&
+		   sscanf(input, "break %x", &address) != 1)
 		{
 		    address = REG_PC;
 		}
